from pathlib import Path

import nhssynth.common.io as io


def check_input_paths(
    dir_experiment: str,
    fn_dataset: str,
    fn_typed: str,
    fn_experiments: str,
    fn_synthetic_datasets: str,
    fn_evaluations: str,
) -> str:
    """
    Sets up the input and output paths for the model files.

    Args:
        dir_experiment: The path to the experiment directory.
        fn_dataset: The base name of the dataset.
        fn_experiments: The filename of the collection of experiments.
        fn_synthetic_datasets: The filename of the collection of synthetic datasets.
        fn_evaluations: The filename of the collection of evaluations.

    Returns:
        The paths
    """
    fn_dataset = Path(fn_dataset).stem
    fn_typed, fn_experiments, fn_synthetic_datasets, fn_evaluations = io.consistent_endings(
        [fn_typed, fn_experiments, fn_synthetic_datasets, fn_evaluations]
    )
    fn_typed, fn_experiments, fn_synthetic_datasets, fn_evaluations = io.potential_suffixes(
        [fn_typed, fn_experiments, fn_synthetic_datasets, fn_evaluations],
        fn_dataset,
    )
    io.warn_if_path_supplied(
        [fn_typed, fn_experiments, fn_synthetic_datasets, fn_evaluations],
        dir_experiment,
    )
    io.check_exists(
        [fn_typed, fn_experiments, fn_synthetic_datasets, fn_evaluations],
        dir_experiment,
    )
    return (
        dir_experiment / fn_typed,
        dir_experiment / fn_experiments,
        dir_experiment / fn_synthetic_datasets,
        dir_experiment / fn_evaluations,
    )



from typing import Optional, Tuple

import numpy as np
import pandas as pd
import torch
from tqdm import tqdm

from nhssynth.modules.model.common.dp import DPMixin
from nhssynth.modules.model.common.mlp import MLP
from nhssynth.modules.model.common.model import Model


class GAN(Model):
    """
    Basic GAN implementation.

    Args:
        n_units_conditional: int
            Number of conditional units
        generator_n_layers_hidden: int
            Number of hidden layers in the generator
        generator_n_units_hidden: int
            Number of hidden units in each layer of the Generator
        generator_activation: string, default 'elu'
            Nonlinearity to use in the generator. Can be 'elu', 'relu', 'selu' or 'leaky_relu'.
        generator_n_iter: int
            Maximum number of iterations in the Generator.
        generator_batch_norm: bool
            Enable/disable batch norm for the generator
        generator_dropout: float
            Dropout value. If 0, the dropout is not used.
        generator_residual: bool
            Use residuals for the generator
        generator_activation_out: Optional[List[Tuple[str, int]]]
            List of activations. Useful with the TabularEncoder
        generator_lr: float = 2e-4
            Generator learning rate, used by the Adam optimizer
        generator_weight_decay: float = 1e-3
            Generator weight decay, used by the Adam optimizer
        generator_opt_betas: tuple = (0.9, 0.999)
            Generator initial decay rates, used by the Adam Optimizer
        generator_extra_penalty_cbks: List[Callable]
            Additional loss callabacks for the generator. Used by the TabularGAN for the conditional loss
        discriminator_n_layers_hidden: int
            Number of hidden layers in the discriminator
        discriminator_n_units_hidden: int
            Number of hidden units in each layer of the discriminator
        discriminator_activation: string, default 'relu'
            Nonlinearity to use in the discriminator. Can be 'elu', 'relu', 'selu' or 'leaky_relu'.
        discriminator_batch_norm: bool
            Enable/disable batch norm for the discriminator
        discriminator_dropout: float
            Dropout value for the discriminator. If 0, the dropout is not used.
        discriminator_lr: float
            Discriminator learning rate, used by the Adam optimizer
        discriminator_weight_decay: float
            Discriminator weight decay, used by the Adam optimizer
        discriminator_opt_betas: tuple
            Initial weight decays for the Adam optimizer
        clipping_value: int, default 0
            Gradients clipping value. Zero disables the feature
        lambda_gradient_penalty: float = 10
            Weight for the gradient penalty
    """

    def __init__(
        self,
        *args,
        noise_dim: int = 128,
        n_units_conditional: int = 0,
        generator_n_layers_hidden: int = 2,
        generator_n_units_hidden: int = 250,
        generator_activation: str = "leaky_relu",
        generator_batch_norm: bool = False,
        generator_dropout: float = 0,
        generator_lr: float = 2e-4,
        generator_residual: bool = True,
        generator_opt_betas: tuple = (0.9, 0.999),
        discriminator_n_layers_hidden: int = 3,
        discriminator_n_units_hidden: int = 300,
        discriminator_activation: str = "leaky_relu",
        discriminator_batch_norm: bool = False,
        discriminator_dropout: float = 0.1,
        discriminator_lr: float = 2e-4,
        discriminator_opt_betas: tuple = (0.9, 0.999),
        clipping_value: int = 0,
        lambda_gradient_penalty: float = 10,
        **kwargs,
    ) -> None:
        super(GAN, self).__init__(*args, **kwargs)

        self.noise_dim = noise_dim
        self.n_units_conditional = n_units_conditional

        self.generator = MLP(
            n_units_in=noise_dim + n_units_conditional,
            n_units_out=self.ncols,
            n_layers_hidden=generator_n_layers_hidden,
            n_units_hidden=generator_n_units_hidden,
            activation=generator_activation,
            batch_norm=generator_batch_norm,
            dropout=generator_dropout,
            lr=generator_lr,
            residual=generator_residual,
            opt_betas=generator_opt_betas,
        ).to(self.device)

        self.discriminator = MLP(
            n_units_in=self.ncols + n_units_conditional,
            n_units_out=1,
            n_layers_hidden=discriminator_n_layers_hidden,
            n_units_hidden=discriminator_n_units_hidden,
            activation=discriminator_activation,
            activation_out=None,
            batch_norm=discriminator_batch_norm,
            dropout=discriminator_dropout,
            lr=discriminator_lr,
            opt_betas=discriminator_opt_betas,
        ).to(self.device)

        self.clipping_value = clipping_value
        self.lambda_gradient_penalty = lambda_gradient_penalty

        def gen_fake_labels(X: torch.Tensor) -> torch.Tensor:
            return torch.zeros((len(X),), device=self.device)

        def gen_true_labels(X: torch.Tensor) -> torch.Tensor:
            return torch.ones((len(X),), device=self.device)

        self.fake_labels_generator = gen_fake_labels
        self.true_labels_generator = gen_true_labels

    @classmethod
    def get_args(cls) -> list[str]:
        return [
            "noise_dim",
            "n_units_conditional",
            "generator_n_layers_hidden",
            "generator_n_units_hidden",
            "generator_activation",
            "generator_batch_norm",
            "generator_dropout",
            "generator_lr",
            "generator_residual",
            "generator_opt_betas",
            "discriminator_n_layers_hidden",
            "discriminator_n_units_hidden",
            "discriminator_activation",
            "discriminator_batch_norm",
            "discriminator_dropout",
            "discriminator_lr",
            "discriminator_opt_betas",
            "clipping_value",
            "lambda_gradient_penalty",
        ]

    @classmethod
    def get_metrics(cls) -> list[str]:
        return ["GLoss", "DLoss"]

    def generate(self, N: int, cond: Optional[np.ndarray] = None) -> np.ndarray:
        N = N or self.nrows
        self.generator.eval()

        condt: Optional[torch.Tensor] = None
        if cond is not None:
            condt = self._check_tensor(cond)
        with torch.no_grad():
            return self.metatransformer.inverse_apply(
                pd.DataFrame(self(N, condt).detach().cpu().numpy(), columns=self.columns)
            )

    def forward(
        self,
        N: int,
        cond: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        if cond is None and self.n_units_conditional > 0:
            # sample from the original conditional
            if self._original_cond is None:
                raise ValueError("Invalid original conditional. Provide a valid value.")
            cond_idxs = torch.randint(len(self._original_cond), (N,))
            cond = self._original_cond[cond_idxs]

        if cond is not None and len(cond.shape) == 1:
            cond = cond.reshape(-1, 1)

        if cond is not None and len(cond) != N:
            raise ValueError("cond length must match N")

        fixed_noise = torch.randn(N, self.noise_dim, device=self.device)
        fixed_noise = self._append_optional_cond(fixed_noise, cond)

        return self.generator(fixed_noise)

    def _train_epoch_generator(
        self,
        X: torch.Tensor,
        cond: Optional[torch.Tensor],
    ) -> float:
        # Update the G network
        self.generator.train()
        self.generator.optim.zero_grad()

        real_X_raw = X.to(self.device)
        real_X = self._append_optional_cond(real_X_raw, cond)
        batch_size = len(real_X)

        noise = torch.randn(batch_size, self.noise_dim, device=self.device)
        noise = self._append_optional_cond(noise, cond)

        fake_raw = self.generator(noise)
        fake = self._append_optional_cond(fake_raw, cond)

        output = self.discriminator(fake).squeeze().float()
        # Calculate G's loss based on this output
        errG = -torch.mean(output)
        if hasattr(self, "generator_extra_penalty_cbks"):
            for extra_loss in self.generator_extra_penalty_cbks:
                errG += extra_loss(
                    real_X_raw,
                    fake_raw,
                    cond=cond,
                )

        # Calculate gradients for G
        errG.backward()

        # Update G
        if self.clipping_value > 0:
            torch.nn.utils.clip_grad_norm_(self.generator.parameters(), self.clipping_value)
        self.generator.optim.step()

        if torch.isnan(errG):
            raise RuntimeError("NaNs detected in the generator loss")

        # Return loss
        return errG.item()

    def _train_epoch_discriminator(
        self,
        X: torch.Tensor,
        cond: Optional[torch.Tensor],
    ) -> float:
        # Update the D network
        self.discriminator.train()

        errors = []

        batch_size = min(self.batch_size, len(X))

        # Train with all-real batch
        real_X = X.to(self.device)
        real_X = self._append_optional_cond(real_X, cond)

        real_labels = self.true_labels_generator(X).to(self.device).squeeze()
        real_output = self.discriminator(real_X).squeeze().float()

        # Train with all-fake batch
        noise = torch.randn(batch_size, self.noise_dim, device=self.device)
        noise = self._append_optional_cond(noise, cond)

        fake_raw = self.generator(noise)
        fake = self._append_optional_cond(fake_raw, cond)

        fake_labels = self.fake_labels_generator(fake_raw).to(self.device).squeeze().float()
        fake_output = self.discriminator(fake.detach()).squeeze()

        # Compute errors. Some fake inputs might be marked as real for privacy guarantees.

        real_real_output = real_output[(real_labels * real_output) != 0]
        real_fake_output = fake_output[(fake_labels * fake_output) != 0]
        errD_real = torch.mean(torch.concat((real_real_output, real_fake_output)))

        fake_real_output = real_output[((1 - real_labels) * real_output) != 0]
        fake_fake_output = fake_output[((1 - fake_labels) * fake_output) != 0]
        errD_fake = torch.mean(torch.concat((fake_real_output, fake_fake_output)))

        penalty = self._loss_gradient_penalty(
            real_samples=real_X,
            fake_samples=fake,
            batch_size=batch_size,
        )
        errD = -errD_real + errD_fake

        self.discriminator.optim.zero_grad()
        if isinstance(self, DPMixin):
            # Adversarial loss
            # 1. split fwd-bkwd on fake and real images into two explicit blocks.
            # 2. no need to compute per_sample_gardients on fake data, disable hooks.
            # 3. re-enable hooks to obtain per_sample_gardients for real data.
            # fake fwd-bkwd
            self.discriminator.disable_hooks()
            penalty.backward(retain_graph=True)
            errD_fake.backward(retain_graph=True)

            self.discriminator.enable_hooks()
            errD_real.backward()  # HACK: calling bkwd without zero_grad() accumulates param gradients
        else:
            penalty.backward(retain_graph=True)
            errD.backward()

        # Update D
        if self.clipping_value > 0:
            torch.nn.utils.clip_grad_norm_(self.discriminator.parameters(), self.clipping_value)
        self.discriminator.optim.step()

        errors.append(errD.item())

        if np.isnan(np.mean(errors)):
            raise RuntimeError("NaNs detected in the discriminator loss")

        return np.mean(errors)

    def _train_epoch(self) -> Tuple[float, float]:
        for data in tqdm(
            self.data_loader,
            desc="Batches",
            position=len(self.stats_bars) + 1,
            leave=False,
        ):
            cond: Optional[torch.Tensor] = None
            if self.n_units_conditional > 0:
                X, cond = data
            else:
                X = data[0]

            losses = {
                "DLoss": self._train_epoch_discriminator(X, cond),
                "GLoss": self._train_epoch_generator(X, cond),
            }
            self._record_metrics(losses)

        return np.mean(self.metrics["GLoss"][-len(self.data_loader) :]), np.mean(
            self.metrics["DLoss"][-len(self.data_loader) :]
        )

    def train(
        self,
        num_epochs: int = 100,
        patience: int = 5,
        displayed_metrics: list[str] = ["GLoss", "DLoss"],
        notebook_run: bool = False,
    ) -> tuple[int, dict[str, np.ndarray]]:
        self._start_training(num_epochs, patience, displayed_metrics, notebook_run)

        for epoch in tqdm(range(num_epochs), desc="Epochs", position=len(self.stats_bars), leave=False):
            losses = self._train_epoch()
            if self._check_patience(epoch, losses[0]) and self._check_patience(epoch, losses[1]):
                num_epochs = epoch + 1
                break

        self._finish_training(num_epochs)
        return (num_epochs, self.metrics)

    def _check_tensor(self, X: torch.Tensor) -> torch.Tensor:
        if isinstance(X, torch.Tensor):
            return X.to(self.device)
        else:
            return torch.from_numpy(np.asarray(X)).to(self.device)

    def _loss_gradient_penalty(
        self,
        real_samples: torch.tensor,
        fake_samples: torch.Tensor,
        batch_size: int,
    ) -> torch.Tensor:
        """Calculates the gradient penalty loss for WGAN GP"""
        # Random weight term for interpolation between real and fake samples
        alpha = torch.rand([batch_size, 1]).to(self.device)
        # Get random interpolation between real and fake samples
        interpolated = (alpha * real_samples + ((1 - alpha) * fake_samples)).requires_grad_(True)
        d_interpolated = self.discriminator(interpolated).squeeze()
        labels = torch.ones((len(interpolated),), device=self.device)

        # Get gradient w.r.t. interpolates
        gradients = torch.autograd.grad(
            outputs=d_interpolated,
            inputs=interpolated,
            grad_outputs=labels,
            create_graph=True,
            retain_graph=True,
            only_inputs=True,
            allow_unused=True,
        )[0]
        gradients = gradients.view(gradients.size(0), -1)
        gradient_penalty = ((gradients.norm(2, dim=-1) - 1) ** 2).mean()
        return self.lambda_gradient_penalty * gradient_penalty

    def _append_optional_cond(self, X: torch.Tensor, cond: Optional[torch.Tensor]) -> torch.Tensor:
        if cond is None:
            return X

        return torch.cat([X, cond], dim=1)



import streamlit as st

from nhssynth.modules.dashboard.utils import hide_streamlit_content

if __name__ == "__main__":
    st.set_page_config(layout="wide")
    hide_streamlit_content()
    if "experiments" not in st.session_state:
        st.error("Upload an evaluation bundle to get started!")
    else:
        st.dataframe(st.session_state["experiments"])



import pathlib
import warnings
from typing import Any, Iterator, Optional, Union

import numpy as np
import pandas as pd
import yaml
from pandas.core.tools.datetimes import _guess_datetime_format_for_array

from nhssynth.common.dicts import filter_dict, get_key_by_value
from nhssynth.modules.dataloader.constraints import ConstraintGraph
from nhssynth.modules.dataloader.missingness import (
    MISSINGNESS_STRATEGIES,
    GenericMissingnessStrategy,
)
from nhssynth.modules.dataloader.transformers import (
    ClusterContinuousTransformer,
    DatetimeTransformer,
    OHECategoricalTransformer,
)
from nhssynth.modules.dataloader.transformers.base import ColumnTransformer


class MetaData:
    class ColumnMetaData:
        def __init__(self, name: str, data: pd.Series, raw: dict) -> None:
            self.name = name
            self.dtype: np.dtype = self._validate_dtype(data, raw.get("dtype"))
            self.categorical: bool = self._validate_categorical(data, raw.get("categorical"))
            self.missingness_strategy: GenericMissingnessStrategy = self._validate_missingness_strategy(
                raw.get("missingness")
            )
            self.transformer: ColumnTransformer = self._validate_transformer(raw.get("transformer"))

        def _validate_dtype(self, data: pd.Series, dtype_raw: Optional[Union[dict, str]] = None) -> np.dtype:
            if isinstance(dtype_raw, dict):
                dtype_name = dtype_raw.pop("name", None)
            elif isinstance(dtype_raw, str):
                dtype_name = dtype_raw
            else:
                dtype_name = self._infer_dtype(data)
            try:
                dtype = np.dtype(dtype_name)
            except TypeError:
                warnings.warn(
                    f"Invalid dtype specification '{dtype_name}' for column '{self.name}', ignoring dtype for this column"
                )
                dtype = self._infer_dtype(data)
            if dtype.kind == "M":
                self._setup_datetime_config(data, dtype_raw)
            elif dtype.kind in ["f", "i", "u"]:
                self.rounding_scheme = self._validate_rounding_scheme(data, dtype, dtype_raw)
            return dtype

        def _infer_dtype(self, data: pd.Series) -> np.dtype:
            return data.dtype.name

        def _infer_datetime_format(self, data: pd.Series) -> str:
            return _guess_datetime_format_for_array(data[data.notna()].astype(str).to_numpy())

        def _setup_datetime_config(self, data: pd.Series, datetime_config: dict) -> dict:
            """
            Add keys to `datetime_config` corresponding to args from the `pd.to_datetime` function
            (see [the docs](https://pandas.pydata.org/docs/reference/api/pandas.to_datetime.html))
            """
            if not isinstance(datetime_config, dict):
                datetime_config = {}
            else:
                datetime_config = filter_dict(datetime_config, {"format", "floor"}, include=True)
            if "format" not in datetime_config:
                datetime_config["format"] = self._infer_datetime_format(data)
            self.datetime_config = datetime_config

        def _validate_rounding_scheme(self, data: pd.Series, dtype: np.dtype, dtype_dict: dict) -> int:
            if dtype_dict and "rounding_scheme" in dtype_dict:
                return dtype_dict["rounding_scheme"]
            else:
                if dtype.kind != "f":
                    return 1.0
                roundable_data = data[data.notna()]
                for i in range(np.finfo(dtype).precision):
                    if (roundable_data.round(i) == roundable_data).all():
                        return 10**-i
            return None

        def _validate_categorical(self, data: pd.Series, categorical: Optional[bool] = None) -> bool:
            if categorical is None:
                return self._infer_categorical(data)
            elif not isinstance(categorical, bool):
                warnings.warn(
                    f"Invalid categorical '{categorical}' for column '{self.name}', ignoring categorical for this column"
                )
                return self._infer_categorical(data)
            else:
                self.boolean = data.nunique() <= 2
                return categorical

        def _infer_categorical(self, data: pd.Series) -> bool:
            self.boolean = data.nunique() <= 2
            return data.nunique() <= 10 or self.dtype.kind == "O"

        def _validate_missingness_strategy(self, missingness_strategy: Optional[Union[dict, str]]) -> tuple[str, dict]:
            if not missingness_strategy:
                return None
            if isinstance(missingness_strategy, dict):
                impute = missingness_strategy.get("impute", None)
                strategy = "impute" if impute else missingness_strategy.get("strategy", None)
            else:
                strategy = missingness_strategy
            if (
                strategy not in MISSINGNESS_STRATEGIES
                or (strategy == "impute" and impute == "mean" and self.dtype.kind != "f")
                or (strategy == "impute" and not impute)
            ):
                warnings.warn(
                    f"Invalid missingness strategy '{missingness_strategy}' for column '{self.name}', ignoring missingness strategy for this column"
                )
                return None
            return (
                MISSINGNESS_STRATEGIES[strategy](impute) if strategy == "impute" else MISSINGNESS_STRATEGIES[strategy]()
            )

        def _validate_transformer(self, transformer: Optional[Union[dict, str]] = {}) -> tuple[str, dict]:
            # if transformer is neither a dict nor a str statement below will raise a TypeError
            if isinstance(transformer, dict):
                self.transformer_name = transformer.get("name")
                self.transformer_config = filter_dict(transformer, "name")
            elif isinstance(transformer, str):
                self.transformer_name = transformer
                self.transformer_config = {}
            else:
                if transformer is not None:
                    warnings.warn(
                        f"Invalid transformer config '{transformer}' for column '{self.name}', ignoring transformer for this column"
                    )
                self.transformer_name = None
                self.transformer_config = {}
            if not self.transformer_name:
                return self._infer_transformer()
            else:
                try:
                    return eval(self.transformer_name)(**self.transformer_config)
                except NameError:
                    warnings.warn(
                        f"Invalid transformer '{self.transformer_name}' or config '{self.transformer_config}' for column '{self.name}', ignoring transformer for this column"
                    )
                    return self._infer_transformer()

        def _infer_transformer(self) -> ColumnTransformer:
            if self.categorical:
                transformer = OHECategoricalTransformer(**self.transformer_config)
            else:
                # Datetime columns should use single Gaussian (typically smooth temporal distributions)
                if self.dtype.kind == "M":
                    datetime_config = self.transformer_config.copy()
                    # Force exactly 1 component for smooth temporal distributions
                    datetime_config["n_components"] = 1
                    transformer = ClusterContinuousTransformer(**datetime_config)
                    transformer = DatetimeTransformer(transformer)
                else:
                    transformer = ClusterContinuousTransformer(**self.transformer_config)
            return transformer

    def __init__(self, data: pd.DataFrame, metadata: Optional[dict] = {}):
        self.columns: pd.Index = data.columns
        self.raw_metadata: dict = metadata
        if set(self.raw_metadata["columns"].keys()) - set(self.columns):
            raise ValueError("Metadata contains keys that do not appear amongst the columns.")
        self.dropped_columns = [cn for cn in self.columns if self.raw_metadata["columns"].get(cn, None) == "drop"]
        self.columns = self.columns.drop(self.dropped_columns)
        self._metadata = {
            cn: self.ColumnMetaData(cn, data[cn], self.raw_metadata["columns"].get(cn, {})) for cn in self.columns
        }
        self.constraints = ConstraintGraph(self.raw_metadata.get("constraints", []), self.columns, self._metadata)

    def __getitem__(self, key: str) -> dict[str, Any]:
        return self._metadata[key]

    def __iter__(self) -> Iterator:
        return iter(self._metadata.values())

    def __repr__(self) -> None:
        return yaml.dump(self._metadata, default_flow_style=False, sort_keys=False)

    @classmethod
    def from_path(cls, data: pd.DataFrame, path_str: str):
        """
        Instantiate a MetaData object from a YAML file via a specified path.

        Args:
            data: The data to be used to infer / validate the metadata.
            path_str: The path to the metadata YAML file.

        Returns:
            The metadata object.
        """
        path = pathlib.Path(path_str)
        if path.exists():
            with open(path) as stream:
                metadata = yaml.safe_load(stream)
            # Filter out the expanded alias/anchor group as it is not needed
            metadata = filter_dict(metadata, {"column_types"})
        else:
            warnings.warn(f"No metadata found at {path}...")
            metadata = {"columns": {}}
        return cls(data, metadata)

    def _collapse(self, metadata: dict) -> dict:
        """
        Given a metadata dictionary, rewrite to collapse duplicate column types in order to leverage YAML anchors and shrink the file.

        Args:
            metadata: The metadata dictionary to be rewritten.

        Returns:
            A rewritten metadata dictionary with collapsed column types and transformers.
                The returned dictionary has the following structure:
                {
                    "column_types": dict,
                    **metadata  # one entry for each column in "columns" that now reference the dicts above
                }
                - "column_types" is a dictionary mapping column type indices to column type configurations.
                - "**metadata" contains the original metadata dictionary, with column types rewritten to use the indices and "column_types".
        """
        c_index = 1
        column_types = {}
        column_type_counts = {}
        for cn, cd in metadata["columns"].items():
            if cd not in column_types.values():
                column_types[c_index] = cd if isinstance(cd, str) else cd.copy()
                column_type_counts[c_index] = 1
                c_index += 1
            else:
                cix = get_key_by_value(column_types, cd)
                column_type_counts[cix] += 1

        for cn, cd in metadata["columns"].items():
            cix = get_key_by_value(column_types, cd)
            if column_type_counts[cix] > 1:
                metadata["columns"][cn] = column_types[cix]
            else:
                column_types.pop(cix)

        return {
            "column_types": {i + 1: x for i, x in enumerate(column_types.values())},
            **metadata,
        }

    def _assemble(self, collapse_yaml: bool) -> dict[str, dict[str, Any]]:
        """
        Rearrange the metadata into a dictionary that can be written to a YAML file.

        Args:
            collapse_yaml: A boolean indicating whether to collapse the YAML representation of the metadata, reducing duplication.

        Returns:
            A dictionary containing the assembled metadata.
        """
        assembled_metadata = {
            "columns": {
                cn: {
                    "dtype": (
                        cmd.dtype.name
                        if not hasattr(cmd, "datetime_config")
                        else {"name": cmd.dtype.name, **cmd.datetime_config}
                    ),
                    "categorical": cmd.categorical,
                }
                for cn, cmd in self._metadata.items()
            }
        }
        # We loop through the base dict above to add other parts if they are present in the metadata
        for cn, cmd in self._metadata.items():
            if cmd.missingness_strategy:
                assembled_metadata["columns"][cn]["missingness"] = (
                    cmd.missingness_strategy.name
                    if cmd.missingness_strategy.name != "impute"
                    else {
                        "name": cmd.missingness_strategy.name,
                        "impute": cmd.missingness_strategy.impute,
                    }
                )
            if cmd.transformer_config:
                assembled_metadata["columns"][cn]["transformer"] = {
                    **cmd.transformer_config,
                    "name": cmd.transformer.__class__.__name__,
                }

        # Add back the dropped_columns not present in the metadata
        if self.dropped_columns:
            assembled_metadata["columns"].update({cn: "drop" for cn in self.dropped_columns})

        if collapse_yaml:
            assembled_metadata = self._collapse(assembled_metadata)

        # We add the constraints section after all of the formatting and processing above
        # In general, the constraints are kept the same as the input (provided they passed validation)
        # If `collapse_yaml` is specified, we output the minimum set of equivalent constraints
        if self.constraints:
            assembled_metadata["constraints"] = (
                [str(c) for c in self.constraints.minimal_constraints]
                if collapse_yaml
                else self.constraints.raw_constraint_strings
            )
        return assembled_metadata

    def save(self, path: pathlib.Path, collapse_yaml: bool) -> None:
        """
        Writes metadata to a YAML file.

        Args:
            path: The path at which to write the metadata YAML file.
            collapse_yaml: A boolean indicating whether to collapse the YAML representation of the metadata, reducing duplication.
        """
        with open(path, "w") as yaml_file:
            yaml.safe_dump(
                self._assemble(collapse_yaml),
                yaml_file,
                default_flow_style=False,
                sort_keys=False,
            )

    def get_sdv_metadata(self) -> dict[str, dict[str, dict[str, str]]]:
        """
        Map combinations of our metadata implementation to SDV's as required by SDMetrics.

        Returns:
            A dictionary containing the SDV metadata.
        """
        sdv_metadata = {
            "columns": {
                cn: {
                    "sdtype": (
                        "boolean"
                        if cmd.boolean
                        else "categorical" if cmd.categorical else "datetime" if cmd.dtype.kind == "M" else "numerical"
                    ),
                }
                for cn, cmd in self._metadata.items()
            }
        }
        for cn, cmd in self._metadata.items():
            if cmd.dtype.kind == "M":
                sdv_metadata["columns"][cn]["format"] = cmd.datetime_config["format"]
        return sdv_metadata

    def save_constraint_graphs(self, path: pathlib.Path) -> None:
        """
        Output the constraint graphs as HTML files.

        Args:
            path: The path at which to write the constraint graph HTML files.
        """
        self.constraints._output_graphs_html(path)



from abc import ABC, abstractmethod
from typing import Optional

import pandas as pd


class ColumnTransformer(ABC):
    """A generic column transformer class to prototype all of the transformers applied via the [`MetaTransformer`][nhssynth.modules.dataloader.metatransformer.MetaTransformer]."""

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def apply(
        self,
        data: pd.DataFrame,
        missingness_column: Optional[pd.Series],
        constraint_adherence: Optional[pd.Series],
    ) -> None:
        """Apply the transformer to the data."""
        pass

    @abstractmethod
    def revert(self, data: pd.DataFrame) -> None:
        """Revert data to pre-transformer state."""
        pass


class TransformerWrapper(ABC):
    """
    A class to facilitate nesting of [`ColumnTransformer`][nhssynth.modules.dataloader.transformers.base.ColumnTransformer]s.

    Args:
        wrapped_transformer: The [`ColumnTransformer`][nhssynth.modules.dataloader.transformers.base.ColumnTransformer] to wrap.
    """

    def __init__(self, wrapped_transformer: ColumnTransformer) -> None:
        super().__init__()
        self._wrapped_transformer: ColumnTransformer = wrapped_transformer

    def apply(
        self,
        data: pd.Series,
        missingness_column: Optional[pd.Series],
        constraint_adherence: Optional[pd.Series],
        **kwargs,
    ) -> pd.DataFrame:
        """Method for applying the wrapped transformer to the data."""
        return self._wrapped_transformer.apply(data, constraint_adherence, missingness_column, **kwargs)

    def revert(self, data: pd.Series, **kwargs) -> pd.DataFrame:
        """Method for reverting the passed data via the wrapped transformer."""
        return self._wrapped_transformer.revert(data, **kwargs)
